@@ -647,7 +647,8 @@ discard_detached_instance (ECalBackendExchangeComponent *ecomp,
 gboolean
 e_cal_backend_exchange_modify_object (ECalBackendExchange *cbex,
 				      icalcomponent *comp,
-				      CalObjModType mod)
+				      CalObjModType mod,
+				      gboolean discard_detached)
 {
 	ECalBackendExchangeComponent *ecomp;
 	const char *uid;
@@ -665,13 +666,16 @@ e_cal_backend_exchange_modify_object (ECalBackendExchange *cbex,
 	if (!ecomp)
 		return FALSE;
 
-	if (mod == CALOBJ_MOD_ALL || icaltime_is_null_time (rid)) {
+	if (mod == CALOBJ_MOD_ALL || icaltime_is_null_time (rid) || discard_detached) {
 		icalcomponent_free (ecomp->icomp);
 		ecomp->icomp = icalcomponent_new_clone (comp);
+		if (discard_detached && !icaltime_is_null_time (rid))
+			discard_detached_instance (ecomp, rid);
 	} else {
-		discard_detached_instance (ecomp, rid);
 		ecomp->instances = g_list_prepend (ecomp->instances,
 						   icalcomponent_new_clone (comp));
+		/*FIXME This is a workaround to cover the timezone issue while setting rid from evolution */
+		discard_detached_instance (ecomp, rid);
 	}
 
 	save_cache (cbex);
@@ -1243,6 +1247,7 @@ set_mode (ECalBackend *backend, CalMode mode)
 {
 	ECalBackendExchange *cbex;
 	ECalBackendExchangePrivate *priv;
+	gboolean re_open = FALSE;
 	
 	cbex = E_CAL_BACKEND_EXCHANGE (backend);
 	priv = cbex->priv;
@@ -1255,7 +1260,10 @@ set_mode (ECalBackend *backend, CalMode mode)
 			cal_mode_to_corba (mode));
 	}
 
+
 	g_mutex_lock (priv->set_lock);	
+	if ((priv->mode == CAL_MODE_LOCAL) && (mode == CAL_MODE_REMOTE))
+		re_open = TRUE;
 
 	switch (mode) {
 	
@@ -1268,7 +1276,7 @@ set_mode (ECalBackend *backend, CalMode mode)
 			priv->mode = CAL_MODE_REMOTE;
 		
 			e_cal_backend_notify_readonly (backend, priv->read_only);
-			if (is_loaded (backend))
+			if (is_loaded (backend) && re_open)
 				e_cal_backend_notify_auth_required(backend);
 			break;
 
